/*******************************************************************/
/* Users  */

/* [Permissions checked!] */

App.Users = new function()
{
	var self = this;

	self.email = null;
	self.fields = null;
	self.field_type_previous = null;
	self.tokens = null;
	self.current_token = null;

	self.selectDialog = function(o)
	{
		if (o.user_id)
		{
			$('#selectUser').val(o.user_id);
		}
		else
		{
			$('#selectUser').val('');
		}

		$('#selectUserForm').unbind('submit');
		$('#selectUserForm').bind('submit', function()
		{
			o.submit({
				id: $('#selectUser').val(),
				name: $('#selectUser option:selected').text()
			});

			return false;
		});

		App.Dialogs.open({
			focusedControl: '#selectUser',
			selectedControl: '#selectUser',
			selector: '#selectUserDialog'
		});
	}

	//---------------------------------------------------------------
	// FIELDS
	//---------------------------------------------------------------

	self.hasField = function(name)
	{
		return self.fields && self.fields.hasOwnProperty(name);
	}

	self.addField = function()
	{
		self._fieldDialog(
		{
			submit: function(o)
			{
				self._loadConfig(
				{
					label: o.label,
					description: o.description,
					name: o.name,
					type_id: o.type_id,
					fallback: o.fallback,

					success: function(html)
					{
						App.Dialogs.closeTop();
						$('#userFields').append(html);
						self._addFieldToArray(o);
					}
				});
			}
		});
	}

	self.addFieldNoDialog = function(o)
	{
		self._loadConfig(
		{
			label: o.label,
			description: o.description,
			name: o.name,
			type_id: o.type_id,
			fallback: o.fallback,

			success: function(html)
			{
				$('#userFields').append(html);
				self._addFieldToArray(o);
			}
		});
	}

	self._addFieldToArray = function(o)
	{
		self.fields[o.name] = {
			label: o.label,
			description: o.description,
			name: o.name,
			type_id: o.type_id,
			fallback: o.fallback
		};
	}

	self._fieldDialog = function(o)
	{
		if (o.field)
		{
			field = self.fields[o.field];
			$('#userFieldLabel').val(field.label);
			$('#userFieldDesc').val(field.description);
			$('#userFieldName').val(field.name);
			$('#userFieldName').attr('disabled', 'disabled');
			$('#userFieldType').attr('disabled', 'disabled');
			$('#userFieldType').val(field.type_id);
			$('#userFieldFallback').val(field.fallback);
			$('#userFieldPassword').val(field.fallback);
		}
		else
		{
			$('#userFieldLabel').val('');
			$('#userFieldDesc').val('');
			$('#userFieldName').val('');
			$('#userFieldName').removeAttr('disabled');
			$('#userFieldType').removeAttr('disabled');
			$('#userFieldType').val('');
			$('#userFieldFallback').val('');
			$('#userFieldPassword').val('');
		}

		self._fieldTypeInit();

		App.Validation.hideErrors();
		$('#userFieldErrors').empty();

		$('#userFieldForm').unbind('submit');
		$('#userFieldForm').submit(function(e)
		{
			App.Validation.hideErrors();
			$('#userFieldErrors').empty();

			var type_id = $('#userFieldType').val();
			var fallback = type_id == 2 ?
				$.trim($('#userFieldPassword').val()) :
				$.trim($('#userFieldFallback').val());

			o.submit(
			{
				label: $.trim($('#userFieldLabel').val()),
				description: $.trim($('#userFieldDesc').val()),
				name: $.trim($('#userFieldName').val()),
				type_id: type_id,
				fallback: fallback
			});

			return false;
		});

		App.Dialogs.open(
		{
			selector: '#userFieldDialog'
		});
	}

	self._fieldTypeInit = function()
	{
		var type_id = $('#userFieldType').val();

		if (type_id == 2)
		{
			$('#userFieldPassword').val($('#userFieldFallback').val());
			$('#userFieldFallback').hide();
			$('#userFieldPassword').show();
			$('#userFieldFallback').val('');
		}
		else
		{
			if (self.field_type_previous == 2)
			{
				$('#userFieldFallback').val($('#userFieldPassword').val());
			}

			$('#userFieldPassword').hide();
			$('#userFieldFallback').show('');
			$('#userFieldPassword').val('');
		}

		self.field_type_previous = type_id;
	}

	self.fieldTypeChanged = function()
	{
		self._fieldTypeInit();
	}

	self._loadConfig = function(o)
	{
		$('#userFieldSubmit').addClass('button-busy');

		App.Ajax.call(
		{
			target: '/admin/integration/ajax_render_user_field',

			arguments:
			{
				label: o.label,
				description: o.description,
				name: o.name,
				type_id: o.type_id,
				fallback: o.fallback
			},

			success: function(data)
			{
				$('#userFieldSubmit').removeClass('button-busy');
				o.success(data);
			},

			error: function(data)
			{
				$('#userFieldSubmit').removeClass('button-busy');
				App.Ajax.handleError(data, '#userFieldErrors');
			}
		});
	}

	self.editField = function(name)
	{
		self._fieldDialog(
		{
			field: name,
			submit: function(o)
			{
				self._loadConfig(
				{
					label: o.label,
					description: o.description,
					name: o.name,
					type_id: o.type_id,
					fallback: o.fallback,

					success: function(html)
					{
						App.Dialogs.closeTop();
						$('#userField-' + name).replaceWith(html);
						self._addFieldToArray(o);
					}
				});
			}
		});
	}

	self.removeField = function(name)
	{
		$('#userField-' + name).remove();
		App.Users._removeFieldFromArray(name);
	}

	self._removeFieldFromArray = function(name)
	{
		delete self.fields[name];
	}

	self.fieldsToString = function()
	{
		var fields = [];

		$.each(self.fields, function(ix, v)
		{
			fields.push(v);
		});

		return JSON.stringify(fields);
	}

	self.editFieldNoDialog = function(o)
	{
		if (!self.hasField(o.name))
		{
			return;
		}

		var field = self.fields[o.name];
		field.fallback = o.fallback; // The only field to update

		self._loadConfig(
		{
			label: field.label,
			description: field.description,
			name: field.name,
			type_id: field.type_id,
			fallback: field.fallback,

			success: function(html)
			{
				$('#userField-' + o.name).replaceWith(html);
				self._addFieldToArray(field);
			}
		});
	}

	//---------------------------------------------------------------
	// TOKENS
	//---------------------------------------------------------------

	self.addToken = function()
	{
		self.current_token = null;

		self._tokenDialog(
		{
			submit: function(o)
			{
				if (!self.current_token)
				{
					self._getToken(o);
				}
				else
				{
					self._addToken(o);
				}
			}
		});
	}

	self._getToken = function(o)
	{
		$('#userTokenGenerate').addClass('button-busy');

		App.Ajax.call(
		{
			target: '/mysettings/ajax_get_api_token',

			arguments:
			{
				name: o.name
			},

			success: function(data)
			{
				$('#userTokenGenerate').removeClass('button-busy');
				$('#userTokenCode').html(data.html).show();
				$('#userTokenGenerate').hide();
				$('#userTokenAdd').show();
				$('#userTokenName').prop('disabled', true);
				self.current_token = data.token;
			},

			error: function(data)
			{
				$('#userTokenGenerate').removeClass('button-busy');
				App.Ajax.handleError(data, '#userTokenErrors');
			}
		});
	}

	self._addToken = function(o)
	{
		$('#userTokenAdd').addClass('button-busy');

		App.Ajax.call(
		{
			target: '/mysettings/ajax_render_api_token',

			arguments:
			{
				name: o.name
			},

			success: function(data)
			{
				$('#userTokenAdd').removeClass('button-busy');
				$('#tokensGrid').append(data.html);
				$('#noTokens').hide();
				App.Dialogs.closeTop();

				self._addTokenToArray(
				{
					token: self.current_token,
					name: data.token.name,
					id: data.token.id
				});
			},

			error: function(data)
			{
				$('#userTokenAdd').removeClass('button-busy');
				App.Ajax.handleError(data, '#userTokenErrors');
			}
		});
	}

	self._addTokenToArray = function(o)
	{
		self.tokens[o.id] = {
			name: o.name,
			token: o.token
		};
	}

	self._tokenDialog = function(o)
	{
		App.Validation.hideErrors();
		$('#userTokenName').prop('disabled', false);
		$('#userTokenName').val('');
		$('#userTokenCode').hide();
		$('#userTokenAdd').hide();
		$('#userTokenGenerate').show();

		$('#userTokenForm').unbind('submit');
		$('#userTokenForm').submit(function(e)
		{
			App.Validation.hideErrors();

			o.submit(
			{
				name: $.trim($('#userTokenName').val())
			});

			return false;
		});

		App.Dialogs.open(
		{
			selector: '#userTokenDialog'
		});
	}

	self.removeToken = function(id)
	{
		$('#userToken-' + id).remove();
		self._removeTokenFromArray(id);
		if ($('#tokensGrid tr.token').length == 0)
		{
			$('#noTokens').show();
		}
	}

	self._removeTokenFromArray = function(id)
	{
		delete self.tokens[id];
	}

	self.tokensToString = function()
	{
		var tokens = [];

		$.each(self.tokens, function(ix, v)
		{
			tokens.push({
				id: ix,
				name: v.name,
				token: v.token
			});
		});

		return JSON.stringify(tokens);
	}

	//---------------------------------------------------------------
	// TOKENS
	//---------------------------------------------------------------

	self.verifyPassword = function()
	{
		self._passwordDialog(
		{
			submit: function(o)
			{
				self._verifyPassword(o);
			}
		});
	}

	self._verifyPassword = function(o)
	{
		$('#userPasswordSubmit').addClass('button-busy');

		App.Ajax.call(
		{
			target: '/mysettings/ajax_verify_password',

			arguments:
			{
				login: o.login,
				password: o.password
			},

			success: function(data)
			{
				$('#userPasswordSubmit').removeClass('button-busy');
				$('#verify_login').val(o.login);
				$('#verify_password').val(o.password);
				App.Dialogs.closeTop();
				$('#form').submit();
			},

			error: function(data)
			{
				$('#userPasswordSubmit').removeClass('button-busy');
				App.Ajax.handleError(data, '#userPasswordErrors');
			}
		});
	}

	self._passwordDialog = function(o)
	{
		App.Validation.hideErrors();
		$('#userPassword').val('');

		$('#userPasswordForm').unbind('submit');
		$('#userPasswordForm').submit(function(e)
		{
			App.Validation.hideErrors();

			var login = null;
			var name = $('#userPasswordName');

			if (name.length)
			{
				login = name.val();
			}
			else
			{
				login = self.email;
			}

			o.submit(
			{
				login: login,
				password: $.trim($('#userPassword').val())
			});

			return false;
		});

		App.Dialogs.open(
		{
			selector: '#userPasswordDialog'
		});
	}

	//---------------------------------------------------------------
	// GOALS
	//---------------------------------------------------------------

	self.hasGoals = function()
	{
		return $('#goals').length > 0;
	}

	self.reloadGoals = function()
	{
		App.Ajax.call(
		{
			target: '/mysettings/ajax_render_goals',
			blockUI: false,

			success: function(html)
			{
				$('#goals').replaceWith(html);
			},

			error: function(data)
			{
				// Ignored, not critical
			}
		});
	}

	self.setGoal = function(step, checked)
	{
		App.Ajax.call(
		{
			target: '/mysettings/ajax_set_goal',

			arguments: {
				step: step,
				checked: checked
			},

			success: function(html)
			{
				$('#goals').replaceWith(html);
			},

			error: function(data)
			{
				App.Ajax.handleError(data);
			}
		});
	}

	self.hideGoals = function(finished)
	{
		App.Ajax.call(
		{
			target: '/mysettings/ajax_hide_goals',
			blockUI: false,

			success: function(html)
			{
				$('#goals-start').remove();
				$('#goals-finish').remove();
				$('#goals-banner').slideUp(
					250,
					function()
					{
						$(this).remove();
						App.Page.reflow();

						if (finished)
						{
							$('#goals-next').show();
						}
						else
						{
							$('#goals').remove();
						}
					}
				);
			},

			error: function(data)
			{
				// Ignored, not critical
			}
		});
	}

	self.hideGoalsNext = function()
	{
		$('#goals').remove();
	}

	$('.searchable').each(function(ix, v) {
		var dropdown = $(v);
		// Apply the chosen control
		dropdown.chosen();
	});

	self.connectAssembla = function() {
		let $asmAppUrlSelector = $('#asmAppUrl');
		let assemblaPortfolio = $asmAppUrlSelector.val() || $asmAppUrlSelector.data('portfolio-url');
		App.Ajax.call({
			target: '/mysettings/ajax_get_assembla_cluster_name',
			arguments: {
				address: assemblaPortfolio
			},
			success: function(data) {
				if (data.result === true) {
					let clusterName = data['x-cluster-name'].toLowerCase();
					Cookies.set('asmclustername', clusterName);
					var $button = $('.asm-connect-btn');
					App.Validation.hideErrors();
					var asmClientId = $button.data(clusterName + 'clientid');
					var apiSubdomain = $button.data(clusterName + 'apisubdomain');
					var authorizationUrl = $button.data('authurl').replace('%s', apiSubdomain).replace('%s', asmClientId)
						+ '&redirect_uri=' + $button.data('redirect-url');

					window.open(authorizationUrl, '', "width=980,height=800");
				}
			},
			error: function(data) {
				App.Validation.hideErrors();
				App.Ajax.handleError(data, '#asmIntegrationErrors');
			},
		});

		return false;
	}

	self.removeAsmClientId = function(removeFromAdmin) {
		var confirm = $('#asmRemoveConfrmMsg').html();
		App.Dialogs.confirm(confirm, function() {
			self.ajaxRemoveAsmClientId(removeFromAdmin);
		});
	}

	self.ajaxRemoveAsmClientId = function(removeFromAdmin) {
		var removeFromAdmin = removeFromAdmin || false;
		App.Ajax.call(
		{
			target: '/mysettings/ajax_remove_client_details',
			arguments: {
				oauth_server: $('.asm-disconnect-btn').data('oauthname'),
			},
			success: function(c) {
				$('.asm-disconnect-btn').hide();
				$('.asm-connect-btn').show();
				$('.user-connected').hide();
				$('.user-notconnected').show();
				$('.asm-oauth-box input').prop('disabled', false);
				$('#asmOauthConnected').val('0');
				$('#asmAppUrl, #asmSpace').val('');
				if (removeFromAdmin) {
					App.Ajax.call({
						target: '/mysettings/ajax_remove_oauth_organization_url',
					});
				}
			},
			error: function(c) {
				App.Ajax.handleError(c, '#asmClientIdErrors');
			}
		});
	}
}
